from skbuild import setup

setup(
    name="llama_cpp_python",
    description="A Python wrapper for llama.cpp",
    version="0.1.1",
    author="Andrei Betlen",
    author_email="abetlen@gmail.com",
    license="MIT",
    packages=['llama_cpp']
)
